###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

import sys

from .HMMMatch import HMMMatch
from .HMMUnit import HMMUnit


class HMMSequence(HMMMatch):
    """
    This class has been adapted from the Perl module written by Genome Research Ltd.
    Perl authors: Rob Finn (rdf@sanger.ac.uk), John Tate (jt6@sanger.ac.uk)
    Perl version: ?
    Python authors: Aaron Mussig (a.mussig@uq.edu.au)
    """

    def __init__(self):
        HMMMatch.__init__(self)
        self.sumEvalue = None
        self.H2mode = None
        self.sumScore = None
        self.desc = None
        self.numberHits = None
        self.exp = None
        self.hmmUnits = list()  # An array of HMMUnit

    def addHMMUnit(self, hmmUnit):
        """
        Adds a hmmUnit to a sequence. It checks that the variable passed in is a HMMUnit object
        """
        if isinstance(hmmUnit, HMMUnit):
            self.hmmUnits.append(hmmUnit)
        else:
            sys.stderr('%s is not a HMMUnit, not added\n.' % hmmUnit)



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################


import gzip
import os
import sys
import traceback

from .exceptions import InputFileError

protein_bases = {'a', 'r', 'n', 'd', 'c', 'q', 'e', 'g', 'h',
                 'i', 'l', 'k', 'm', 'f', 'p', 's', 't', 'w', 'y', 'v'}
nucleotide_bases = {'a', 'c', 'g', 't'}
insertion_bases = {'-', '.'}


def read_fasta(fasta_file, keep_annotation=False):
    """Read sequences from fasta file.

    Parameters
    ----------
    fasta_file : str
        Name of fasta file to read.
    keep_annotation : boolean
        Determine is sequence id should contain annotation.

    Returns
    -------
    dict : dict[seq_id] -> seq
        Sequences indexed by sequence id.
    """

    if not os.path.exists(fasta_file):
        raise InputFileError('Input file %s does not exist.' % fasta_file)

    if os.stat(fasta_file).st_size == 0:
        return {}

    try:

        if fasta_file.endswith('.gz'):
            file_f, file_mode = gzip.open, 'rt'
        else:
            file_f, file_mode = open, 'r'

        seqs = {}
        with file_f(fasta_file, file_mode) as f:

            for line in f.readlines():
                # skip blank lines
                if not line.strip():
                    continue

                if line[0] == '>':
                    if keep_annotation:
                        seq_id = line[1:-1]
                    else:
                        seq_id = line[1:].split(None, 1)[0]

                    seqs[seq_id] = []
                else:
                    seqs[seq_id].append(line.strip())

        for seq_id, seq in seqs.items():
            seqs[seq_id] = ''.join(seq).replace(' ', '')
    except:
        print(traceback.format_exc())
        print()
        print("[Error] Failed to process sequence file: " + fasta_file)
        sys.exit(1)

    return seqs


def read_fasta_seq(fasta_file, keep_annotation=False):
    """Generator function to read sequences from fasta file.

    This function is intended to be used as a generator
    in order to avoid having to have large sequence files
    in memory. Input file may be gzipped.

    Example:
    seq_io = SeqIO()
    for seq_id, seq in seq_io.read_fasta_seq(fasta_file):
        print seq_id
        print seq

    Parameters
    ----------
    fasta_file : str
        Name of fasta file to read.
    keep_annotation : boolean
        Determine if annotation string should be returned.

    Yields
    ------
    list : [seq_id, seq, [annotation]]
        Unique id of the sequence followed by the sequence itself,
        and the annotation if keep_annotation is True.
    """

    if not os.path.exists(fasta_file):
        raise InputFileError('Input file %s does not exist.' % fasta_file)

    if os.stat(fasta_file).st_size == 0:
        pass

    try:
        open_file = open
        if fasta_file.endswith('.gz'):
            open_file = gzip.open

        seq_id = None
        annotation = None
        seq = None
        with open_file(fasta_file, 'r') as f:

            for line in f.readlines():
                # skip blank lines
                if not line.strip():
                    continue

                if line[0] == '>':
                    if seq_id is not None:
                        if keep_annotation:
                            yield seq_id, ''.join(seq).replace(' ', ''), annotation
                        else:
                            yield seq_id, ''.join(seq).replace(' ', '')

                    line_split = line[1:-1].split(None, 1)
                    if len(line_split) == 2:
                        seq_id, annotation = line_split
                    else:
                        seq_id = line_split[0]
                        annotation = ''
                    seq = []
                else:
                    seq.append(line.strip())

        # report last sequence
        if seq:
            if keep_annotation:
                yield seq_id, ''.join(seq).replace(' ', ''), annotation
            else:
                yield seq_id, ''.join(seq).replace(' ', '')
    except GeneratorExit:
        pass
    except:
        print(traceback.format_exc())
        print()
        print("[Error] Failed to process sequence file: " + fasta_file)
        sys.exit(1)


def read_seq(seq_file, keep_annotation=False):
    """Generator function to read sequences from fasta/q file.

    This function is intended to be used as a generator
    in order to avoid having to have large sequence files
    in memory. Input file may be gzipped and in either
    fasta or fastq format. It is slightly more efficient
    to directly call read_fasta_seq() or read_fastq_seq()
    if the type of input file in known.

    Example:
    seq_io = SeqIO()
    for seq_id, seq in seq_io.read_seq(fasta_file):
        print seq_id
        print seq

    Parameters
    ----------
    seq_file : str
        Name of fasta/q file to read.
    keep_annotation : boolean
        Determine if annotation string should be returned.

    Yields
    ------
    list : [seq_id, seq, [annotation]]
        Unique id of the sequence followed by the sequence itself,
        and the annotation if keep_annotation is True.
    """

    if seq_file.endswith(('.fq.gz', '.fastq.gz', '.fq', '.fq.gz')):
        for rtn in read_fastq_seq(seq_file):
            yield rtn
    else:
        for rtn in read_fasta_seq(seq_file, keep_annotation):
            yield rtn


def write_fasta(seqs, fasta_file, wrap=80):
    """Write sequences to a fasta file.

    Parameters
    ----------
    seqs : dict[seq_id] -> seq
        Sequences indexed by sequence id.
    fasta_file : str
        Path to write the sequences to.
    wrap: int
        Number of AA/NT before the line is wrapped.
    """
    with open(fasta_file, 'w') as f:
        for gid, gseq in seqs.items():
            f.write('>{}\n'.format(gid))
            for i in range(0, len(gseq), wrap):
                f.write('{}\n'.format(gseq[i:i + wrap]))



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

import sys

from .HMMSequence import HMMSequence
from .HMMUnit import HMMUnit


class HMMResults(object):
    """
     This class has been adapted from the Perl module written by Genome Research Ltd.
     Perl authors: Rob Finn (rdf@sanger.ac.uk)
     Perl version: HMMResults.pm,v 1.3 2009-12-15 14:38:08 jt6 Exp $
     Python authors: Aaron Mussig (a.mussig@uq.edu.au)
     """

    def __init__(self):
        self.hmmerVersion = None
        self.hmmName = None
        self.seqDB = None
        self.hmmLength = None
        self.thisFile = None
        self.seedName = None
        self.seqs = dict()
        self.units = list()
        self.domThr = 25.0
        self.seqThr = 25.0
        self.evalueThr = None
        self.domTC = None
        self.seqTC = None
        self.domNC = None
        self.seqNC = None
        self.randSeedNum = None
        self.description = None
        self.seqName = None
        self.seqLength = None
        self.eof = False
        self.program = None

    def addHMMSeq(self, hmmSeq):
        """
        Adds a HMMSequence object to the results object
        """
        if not isinstance(hmmSeq, HMMSequence):
            sys.exit('Trying to add a non HMMSequence object')

        if hmmSeq.name in self.seqs:
            sys.exit('Trying to add the same sequence twice')

        self.seqs[hmmSeq.name] = hmmSeq

    def addHMMUnit(self, hmmUnit):
        """
        Adds HMM units (the actual region hit) to the HMMSequence in the object and for convenience to
        the results sets. All we store are duplicates of the references.
        """
        if not isinstance(hmmUnit, HMMUnit):
            sys.exit('Trying to add a non HMMUnit object')

        if self.seqs:
            if hmmUnit.name in self.seqs:
                self.seqs[hmmUnit.name].addHMMUnit(hmmUnit)
            else:
                sys.stderr('Could not add hmmUnit as the sequence has not been added\n')

        # More conveinence we store the point to the hmmunit in an array
        self.units.append(hmmUnit)

    def remove_overlaps_by_clan(self, clanmap, nested):
        new = HMMResults()
        new.seqName = self.seqName

        for unit in sorted(self.units, key=lambda x: float(x.evalue), reverse=False):

            # check if it overlaps before adding
            o = None

            for u in new.units:
                if unit.name in clanmap and u.name in clanmap and clanmap[unit.name] == clanmap[u.name]:
                    if overlap(unit, u):
                        # TODO: Did i interpret the logic correctly here?
                        if unit.name in nested and u.name in nested[unit.name]:
                            continue
                        else:
                            o = True
                            break

            if not o:
                if unit.name not in new.seqs:
                    new_seq = HMMSequence()
                    new_seq.name = self.seqs[unit.name].name
                    new_seq.desc = self.seqs[unit.name].desc
                    new_seq.bits = self.seqs[unit.name].bits
                    new_seq.evalue = self.seqs[unit.name].evalue
                    new_seq.numberHits = self.seqs[unit.name].numberHits

                    new.addHMMSeq(new_seq)

                new.addHMMUnit(unit)

        return new


def overlap(unit1, unit2):
    """
    does unit1 overlap with unit2?
    """
    out = sorted([unit1, unit2], key=lambda x: int(x.seqFrom), reverse=False)
    u1 = out[0]
    u2 = out[1]

    if u2.seqFrom <= u1.seqTo:
        return True

    return False



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

import logging
import os
import subprocess

from gtdbtk.biolib_lite.common import make_sure_path_exists
from gtdbtk.biolib_lite.execute import check_dependencies
from gtdbtk.exceptions import FastTreeException


class FastTree(object):
    """Python wrapper for FastTree: http://www.microbesonline.org/fasttree/"""

    def __init__(self):
        """Instantiate the class."""
        self.logger = logging.getLogger('timestamp')
        self.version = self.get_version()

    def get_version(self):
        """ Get FastTree version. """
        try:
            env = os.environ.copy()
            args = ['FastTree']
            proc = subprocess.Popen(args, stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE, env=env, encoding='utf-8')
            output, error = proc.communicate()
            return error.split('\n')[0].split(' ')[4]
        except:
            return "(version unavailable)"

    def run(self, output_tree, tree_log, fasttree_log, prot_model, no_support, no_gamma, msa_file, cpus=1):
        """Run FastTree.

        Parameters
        ----------
        output_tree : str
            The path where the resulting tree should be written to.
        tree_log : str
            The path where the FastTree stats should be written to.
        fasttree_log : str
            The path where the FastTree log should be written to.
        prot_model : str
            Either 'JTT', 'WAG', or 'LG'.
        no_support : bool
            True if no support should be used, False otherwise.
        no_gamma : bool
            True if no gamma should be used, False otherwise.
        msa_file : str
            The path to the input MSA.
        cpus : int
            The maximum number of CPUs for FastTree to use.

        Raises
        ------
        FastTreeException
            If an error is encountered while running FastTree.

        """
        env = os.environ.copy()
        if cpus > 1:
            cmd = 'FastTreeMP'
            env['OMP_NUM_THREADS'] = str(cpus)
        else:
            cmd = 'FastTree'
        check_dependencies([cmd])

        make_sure_path_exists(os.path.dirname(output_tree))
        make_sure_path_exists(os.path.dirname(tree_log))
        make_sure_path_exists(os.path.dirname(fasttree_log))

        # Setup arguments
        args = [cmd]
        if prot_model == 'WAG':
            args.append('-wag')
        elif prot_model == 'LG':
            args.append('-lg')
        if no_support:
            args.append('-nosupport')
        if not no_gamma:
            args.append('-gamma')
        args.append('-log')
        args.append(tree_log)
        args.append(msa_file)

        model_out = [prot_model,
                     ('-' if no_gamma else '+') + 'gamma',
                     ('no' if no_support else '') + 'support']
        self.logger.info('Inferring FastTree ({}) using a maximum of {} CPUs.'.format(
            ', '.join(model_out), cpus))
        self.logger.info('FastTree version: {}'.format(self.version))

        with open(output_tree, 'w') as f_out_tree:
            with open(fasttree_log, 'w') as f_out_err:
                proc = subprocess.Popen(
                    args, stdout=f_out_tree, stderr=f_out_err, env=env)
                proc.communicate()

        # Validate results
        if proc.returncode != 0:
            self.logger.error(
                'An error was encountered while running FastTree.')
            raise FastTreeException('FastTree returned a non-zero exit code.')
        if not os.path.isfile(output_tree):
            self.logger.error(
                'An error was encountered while running FastTree.')
            raise FastTreeException(
                'Tree output file is missing: {}'.format(output_tree))
        elif os.path.getsize(output_tree) < 1:
            self.logger.error(
                'An error was encountered while running FastTree.')
            raise FastTreeException(
                'Tree output file is empty: {}'.format(output_tree))



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

import logging
import os
from collections import defaultdict
from shutil import copy

import numpy as np

import gtdbtk.config.config as Config
from gtdbtk.biolib_lite.execute import check_dependencies
from gtdbtk.biolib_lite.seq_io import read_fasta
from gtdbtk.biolib_lite.taxonomy import Taxonomy
from gtdbtk.config.output import *
from gtdbtk.exceptions import GenomeMarkerSetUnknown, MSAMaskLengthMismatch
from gtdbtk.external.hmm_aligner import HmmAligner
from gtdbtk.external.pfam_search import PfamSearch
from gtdbtk.external.prodigal import Prodigal
from gtdbtk.external.tigrfam_search import TigrfamSearch
from gtdbtk.tools import merge_two_dicts, symlink_f
from gtdbtk.trim_msa import TrimMSA


class Markers(object):
    """Identify and align marker genes."""

    def __init__(self, cpus=1):
        """Initialize."""

        self.logger = logging.getLogger('timestamp')
        self.warnings = logging.getLogger('warnings')

        self.cpus = cpus

        self.genome_file_suffix = GENOME_FILE_SUFFIX
        self.protein_file_suffix = PROTEIN_FILE_SUFFIX
        self.nt_gene_file_suffix = NT_GENE_FILE_SUFFIX
        self.gff_file_suffix = GFF_FILE_SUFFIX
        self.checksum_suffix = CHECKSUM_SUFFIX

        self.taxonomy_file = Config.TAXONOMY_FILE

        self.pfam_hmm_dir = Config.PFAM_HMM_DIR
        self.pfam_suffix = PFAM_SUFFIX
        self.pfam_top_hit_suffix = PFAM_TOP_HIT_SUFFIX

        self.tigrfam_hmms = Config.TIGRFAM_HMMS
        self.tigrfam_suffix = TIGRFAM_SUFFIX
        self.tigrfam_top_hit_suffix = TIGRFAM_TOP_HIT_SUFFIX

    def _report_identified_marker_genes(self, gene_dict, outdir, prefix):
        """Report statistics for identified marker genes."""

        translation_table_file = open(os.path.join(
            outdir, PATH_TLN_TABLE_SUMMARY.format(prefix=prefix)), "w")
        bac_outfile = open(os.path.join(
            outdir, PATH_BAC120_MARKER_SUMMARY.format(prefix=prefix)), "w")
        arc_outfile = open(os.path.join(
            outdir, PATH_AR122_MARKER_SUMMARY.format(prefix=prefix)), "w")

        header = "Name\tnumber_unique_genes\tnumber_multiple_genes\tnumber_missing_genes\tlist_unique_genes\tlist_multiple_genes\tlist_missing_genes\n"

        bac_outfile.write(header)
        arc_outfile.write(header)

        # gather information for all marker genes
        marker_dbs = {"PFAM": PFAM_TOP_HIT_SUFFIX,
                      "TIGR": TIGRFAM_TOP_HIT_SUFFIX}

        marker_bac_list_original = []
        for db_marker in Config.BAC120_MARKERS.keys():
            marker_bac_list_original.extend([marker.replace(".HMM", "").replace(".hmm", "")
                                             for marker in Config.BAC120_MARKERS[db_marker]])

        marker_arc_list_original = []
        for db_marker in Config.AR122_MARKERS.keys():
            marker_arc_list_original.extend([marker.replace(".HMM", "").replace(".hmm", "")
                                             for marker in Config.AR122_MARKERS[db_marker]])

        for db_genome_id, info in gene_dict.items():

            unique_genes_bac, multi_hits_bac, missing_genes_bac = [], [], []
            unique_genes_arc, multi_hits_arc, missing_genes_arc = [], [], []

            gene_bac_dict, gene_arc_dict = {}, {}

            path = info.get("aa_gene_path")
            for _marker_db, marker_suffix in marker_dbs.items():
                # get all gene sequences
                protein_file = str(path)
                tophit_path = protein_file.replace(
                    PROTEIN_FILE_SUFFIX, marker_suffix)

                # we load the list of all the genes detected in the genome
                all_genes_dict = read_fasta(protein_file, False)

                # Prodigal adds an asterisks at the end of each called genes.
                # These asterisks sometimes appear in the MSA, which can be
                # an issue for some downstream software
                for seq_id, seq in all_genes_dict.items():
                    if seq[-1] == '*':
                        all_genes_dict[seq_id] = seq[:-1]

                # we store the tophit file line by line and store the
                # information in a dictionary
                with open(tophit_path) as tp:
                    # first line is header line
                    tp.readline()

                    for line_tp in tp:
                        linelist = line_tp.split("\t")
                        genename = linelist[0]
                        sublist = linelist[1]
                        if ";" in sublist:
                            diff_markers = sublist.split(";")
                        else:
                            diff_markers = [sublist]

                        for each_mark in diff_markers:
                            sublist = each_mark.split(",")
                            markerid = sublist[0]

                            if (markerid not in marker_bac_list_original and
                                    markerid not in marker_arc_list_original):
                                continue

                            if markerid in marker_bac_list_original:
                                if markerid in gene_bac_dict:
                                    gene_bac_dict.get(markerid)[
                                        "multihit"] = True
                                else:
                                    gene_bac_dict[markerid] = {
                                        "gene": genename,
                                        "multihit": False}

                            if markerid in marker_arc_list_original:
                                if markerid in gene_arc_dict:
                                    gene_arc_dict.get(markerid)[
                                        "multihit"] = True
                                else:
                                    gene_arc_dict[markerid] = {
                                        "gene": genename,
                                        "multihit": False}

            for mid in marker_bac_list_original:
                if mid not in gene_bac_dict:
                    missing_genes_bac.append(mid)
                elif gene_bac_dict[mid]["multihit"]:
                    multi_hits_bac.append(mid)
                else:
                    unique_genes_bac.append(mid)

            for mid in marker_arc_list_original:
                if mid not in gene_arc_dict:
                    missing_genes_arc.append(mid)
                elif gene_arc_dict[mid]["multihit"]:
                    multi_hits_arc.append(mid)
                else:
                    unique_genes_arc.append(mid)

            bac_outfile.write("{0}\t{1}\t{2}\t{3}\t{4}\t{5}\t{6}\n".format(db_genome_id,
                                                                           len(unique_genes_bac),
                                                                           len(multi_hits_bac),
                                                                           len(missing_genes_bac),
                                                                           ','.join(
                                                                               unique_genes_bac),
                                                                           ','.join(
                                                                               multi_hits_bac),
                                                                           ','.join(missing_genes_bac)))

            arc_outfile.write("{0}\t{1}\t{2}\t{3}\t{4}\t{5}\t{6}\n".format(db_genome_id,
                                                                           len(unique_genes_arc),
                                                                           len(multi_hits_arc),
                                                                           len(missing_genes_arc),
                                                                           ','.join(
                                                                               unique_genes_arc),
                                                                           ','.join(
                                                                               multi_hits_arc),
                                                                           ','.join(missing_genes_arc)))

            translation_table_file.write('{}\t{}\n'.format(
                db_genome_id, info.get("best_translation_table")))

        bac_outfile.close()
        arc_outfile.close()
        translation_table_file.close()

        # Create a symlink to store the summary files in the root.
        symlink_f(PATH_BAC120_MARKER_SUMMARY.format(prefix=prefix),
                  os.path.join(outdir, os.path.basename(PATH_BAC120_MARKER_SUMMARY.format(prefix=prefix))))
        symlink_f(PATH_AR122_MARKER_SUMMARY.format(prefix=prefix),
                  os.path.join(outdir, os.path.basename(PATH_AR122_MARKER_SUMMARY.format(prefix=prefix))))
        symlink_f(PATH_TLN_TABLE_SUMMARY.format(prefix=prefix),
                  os.path.join(outdir, os.path.basename(PATH_TLN_TABLE_SUMMARY.format(prefix=prefix))))

    def identify(self, genomes, out_dir, prefix, force):
        """Identify marker genes in genomes.

        Parameters
        ----------
        genomes : dict
            Genome IDs as the key, path to genome file as value.
        out_dir : str
            Path to the output directory.
        prefix : str
            Prefix to append to generated files.
        force : bool
            Overwrite any existing files.

        Raises
        ------
        GTDBTkException
            If an exception is encountered during the identify step.

        """
        check_dependencies(['prodigal', 'hmmsearch'])

        self.logger.info('Identifying markers in %d genomes with %d threads.' % (len(genomes),
                                                                                 self.cpus))

        self.marker_gene_dir = os.path.join(out_dir, DIR_MARKER_GENE)
        prodigal = Prodigal(self.cpus,
                            False,
                            self.marker_gene_dir,
                            self.protein_file_suffix,
                            self.nt_gene_file_suffix,
                            self.gff_file_suffix,
                            force)
        self.logger.info(
            "Running Prodigal {} to identify genes.".format(prodigal.version))
        genome_dictionary = prodigal.run(genomes)

        # annotated genes against TIGRFAM and Pfam databases
        self.logger.info("Identifying TIGRFAM protein families.")
        gene_files = [genome_dictionary[db_genome_id]['aa_gene_path']
                      for db_genome_id in genome_dictionary.keys()]

        tigr_search = TigrfamSearch(self.cpus,
                                    self.tigrfam_hmms,
                                    self.protein_file_suffix,
                                    self.tigrfam_suffix,
                                    self.tigrfam_top_hit_suffix,
                                    self.checksum_suffix,
                                    self.marker_gene_dir)
        tigr_search.run(gene_files)

        self.logger.info("Identifying Pfam protein families.")
        pfam_search = PfamSearch(self.cpus,
                                 self.pfam_hmm_dir,
                                 self.protein_file_suffix,
                                 self.pfam_suffix,
                                 self.pfam_top_hit_suffix,
                                 self.checksum_suffix,
                                 self.marker_gene_dir)
        pfam_search.run(gene_files)
        self.logger.info(
            "Annotations done using HMMER {}".format(tigr_search.version))

        self._report_identified_marker_genes(
            genome_dictionary, out_dir, prefix)

    def _path_to_identify_data(self, identity_dir, warn=True):
        """Get path to genome data produced by 'identify' command."""

        marker_gene_dir = os.path.join(identity_dir, DIR_MARKER_GENE)

        genomic_files = {}
        lq_gids = list()
        for gid in os.listdir(marker_gene_dir):
            gid_dir = os.path.join(marker_gene_dir, gid)
            if not os.path.isdir(gid_dir):
                continue

            aa_gene_path = os.path.join(
                gid_dir, gid + self.protein_file_suffix)

            # Check if any genes were called
            if os.path.getsize(aa_gene_path) < 1:
                lq_gids.append(gid)
            else:
                genomic_files[gid] = {'aa_gene_path': aa_gene_path,
                                      'translation_table_path': os.path.join(gid_dir, 'prodigal' + TRANSLATION_TABLE_SUFFIX),
                                      'nt_gene_path': os.path.join(gid_dir, gid + self.nt_gene_file_suffix),
                                      'gff_path': os.path.join(gid_dir, gid + self.gff_file_suffix)
                                      }

        if len(lq_gids) > 0 and warn:
            self.logger.warning(f'Excluding {len(lq_gids)} genomes '
                                f'in the identify directory which have no genes '
                                f'called (see gtdb.warnings.log)')
            self.warnings.warning(f'Excluding the following {len(lq_gids)} genomes '
                                  f'which were found in the identify directory '
                                  f'with no genes called.')
            for lq_gid in lq_gids:
                self.warnings.info(lq_gid)
        return genomic_files

    def _msa_filter_by_taxa(self, concatenated_file, gtdb_taxonomy, taxa_filter, outgroup_taxon):
        """Filter GTDB MSA filtered to specified taxa."""

        msa = read_fasta(concatenated_file)
        msa_len = len(msa)
        self.logger.info(
            'Read concatenated alignment for %d GTDB genomes.' % msa_len)

        if taxa_filter is not None:
            taxa_to_keep = set(taxa_filter.split(','))

            if outgroup_taxon not in taxa_to_keep and outgroup_taxon is not None:
                taxa_to_keep.add(outgroup_taxon)

            filtered_genomes = 0
            for genome_id, taxa in gtdb_taxonomy.items():
                common_taxa = taxa_to_keep.intersection(taxa)
                if len(common_taxa) == 0:
                    if genome_id in msa:
                        del msa[genome_id]
                        filtered_genomes += 1

            msg = 'Filtered %.2f%% (%d/%d) taxa based on assigned taxonomy, ' \
                  '%d taxa remain.' % (
                      (float(filtered_genomes) / float(msa_len)) * 100.0,
                      filtered_genomes, msa_len, msa_len - filtered_genomes)
            self.logger.info(msg) if len(msa) > 0 else self.logger.warning(msg)

        return msa

    def _apply_mask(self, gtdb_msa, user_msa, msa_mask, min_perc_aa):
        """Apply canonical mask to MSA file."""
        aligned_genomes = merge_two_dicts(gtdb_msa, user_msa)
        list_mask = np.fromfile(msa_mask, dtype='S1') == b'1'

        output_seqs = {}
        pruned_seqs = {}
        for seq_id, seq in aligned_genomes.items():
            list_seq = np.fromiter(seq, dtype='S1')
            if list_mask.shape[0] != list_seq.shape[0]:
                raise MSAMaskLengthMismatch(
                    'Mask and alignment length do not match.')

            list_masked_seq = list_seq[list_mask]

            masked_seq_unique = np.unique(list_masked_seq, return_counts=True)
            masked_seq_counts = defaultdict(lambda: 0)
            for aa_char, aa_count in zip(masked_seq_unique[0], masked_seq_unique[1]):
                masked_seq_counts[aa_char.decode('utf-8')] = aa_count

            masked_seq = list_masked_seq.tostring().decode('utf-8')

            valid_bases = list_masked_seq.shape[0] - \
                masked_seq_counts['.'] - masked_seq_counts['-']
            if seq_id in user_msa and valid_bases < list_masked_seq.shape[0] * min_perc_aa:
                pruned_seqs[seq_id] = masked_seq
                continue

            output_seqs[seq_id] = masked_seq

        return output_seqs, pruned_seqs

    def _write_msa(self, seqs, output_file, gtdb_taxonomy):
        """Write sequences to FASTA file."""

        with open(output_file, 'w') as fout:
            for genome_id, alignment in seqs.items():
                if genome_id in gtdb_taxonomy:
                    fout.write('>%s %s\n' %
                               (genome_id, ';'.join(gtdb_taxonomy[genome_id])))
                else:
                    fout.write('>%s\n' % genome_id)
                fout.write('%s\n' % alignment)

    def genome_domain(self, identity_dir, prefix):
        """Determine domain of User genomes based on identified marker genes."""

        bac_count = defaultdict(int)
        ar_count = defaultdict(int)

        for d, marker_file in ((bac_count, os.path.join(identity_dir, PATH_BAC120_MARKER_SUMMARY.format(prefix=prefix))),
                               (ar_count, os.path.join(identity_dir, PATH_AR122_MARKER_SUMMARY.format(prefix=prefix)))):
            with open(marker_file) as f:
                f.readline()

                for line in f:
                    line_split = line.strip().split('\t')

                    gid = line_split[0]
                    num_markers = int(line_split[1])

                    d[gid] = num_markers

        bac_gids = set()
        ar_gids = set()
        bac_ar_diff = {}
        for gid in bac_count:
            arc_aa_per = (ar_count[gid] * 100.0 / Config.AR_MARKER_COUNT)
            bac_aa_per = (bac_count[gid] * 100.0 / Config.BAC_MARKER_COUNT)
            if bac_aa_per >= arc_aa_per:
                bac_gids.add(gid)
            else:
                ar_gids.add(gid)
            if abs(bac_aa_per - arc_aa_per) <= 10:
                bac_ar_diff[gid] = {'bac120': round(
                    bac_aa_per, 1), 'ar122': round(arc_aa_per, 1)}

        return bac_gids, ar_gids, bac_ar_diff

    def _write_marker_info(self, marker_db, marker_file):
        """Write out information about markers comprising MSA."""

        marker_paths = {"PFAM": os.path.join(self.pfam_hmm_dir, 'individual_hmms'),
                        "TIGRFAM": os.path.join(os.path.dirname(self.tigrfam_hmms), 'individual_hmms')}

        with open(marker_file, 'w') as fh:
            fh.write('Marker ID\tName\tDescription\tLength (bp)\n')
            for db_marker in sorted(marker_db):
                for marker in marker_db[db_marker]:
                    marker_id = marker[0:marker.rfind('.')]
                    marker_path = os.path.join(marker_paths[db_marker], marker)

                    # get marker name, description, and size
                    with open(marker_path) as fp:
                        for line in fp:
                            if line.startswith("NAME  "):
                                marker_name = line.split("  ")[1].strip()
                            elif line.startswith("DESC  "):
                                marker_desc = line.split("  ")[1].strip()
                            elif line.startswith("LENG  "):
                                marker_size = line.split("  ")[1].strip()
                                break
                    fh.write('%s\t%s\t%s\t%s\n' %
                             (marker_id, marker_name, marker_desc, marker_size))

    def align(self,
              identify_dir,
              skip_gtdb_refs,
              taxa_filter,
              min_perc_aa,
              custom_msa_filters,
              skip_trimming,
              rnd_seed,
              cols_per_gene,
              min_consensus,
              max_consensus,
              min_per_taxa,
              out_dir,
              prefix,
              outgroup_taxon):
        """Align marker genes in genomes."""

        if identify_dir != out_dir:
            if not os.path.isdir(os.path.join(out_dir, DIR_IDENTIFY)):
                os.makedirs(os.path.join(out_dir, DIR_IDENTIFY))

            copy(os.path.join(identify_dir, PATH_BAC120_MARKER_SUMMARY.format(prefix=prefix)),
                 os.path.join(out_dir, DIR_IDENTIFY))
            copy(os.path.join(identify_dir, PATH_AR122_MARKER_SUMMARY.format(prefix=prefix)),
                 os.path.join(out_dir, DIR_IDENTIFY))

            identify_gene_file = os.path.join(
                identify_dir, PATH_TLN_TABLE_SUMMARY.format(prefix=prefix))
            copy(identify_gene_file, os.path.join(out_dir, DIR_IDENTIFY))

        if not os.path.exists(os.path.join(out_dir, DIR_ALIGN_INTERMEDIATE)):
            os.makedirs(os.path.join(out_dir, DIR_ALIGN_INTERMEDIATE))

        # write out files with marker information
        bac120_marker_info_file = os.path.join(
            out_dir, PATH_BAC120_MARKER_INFO.format(prefix=prefix))
        self._write_marker_info(Config.BAC120_MARKERS, bac120_marker_info_file)
        ar122_marker_info_file = os.path.join(
            out_dir, PATH_AR122_MARKER_INFO.format(prefix=prefix))
        self._write_marker_info(Config.AR122_MARKERS, ar122_marker_info_file)

        genomic_files = self._path_to_identify_data(
            identify_dir, identify_dir != out_dir)
        self.logger.info('Aligning markers in %d genomes with %d threads.' % (len(genomic_files),
                                                                              self.cpus))

        # determine marker set for each user genome
        bac_gids, ar_gids, _bac_ar_diff = self.genome_domain(
            identify_dir, prefix)

        # align user genomes
        gtdb_taxonomy = Taxonomy().read(self.taxonomy_file)
        for gids, msa_file, mask_file, marker_set_id in ((bac_gids, Config.CONCAT_BAC120, Config.MASK_BAC120, "bac120"),
                                                         (ar_gids, Config.CONCAT_AR122, Config.MASK_AR122, "ar122")):

            if len(gids) == 0:
                continue

            if marker_set_id == 'bac120':
                self.logger.info(
                    'Processing %d genomes identified as bacterial.' % len(gids))
                marker_info_file = bac120_marker_info_file
                marker_filtered_genomes = os.path.join(
                    out_dir, PATH_BAC120_FILTERED_GENOMES.format(prefix=prefix))
                marker_msa_path = os.path.join(
                    out_dir, PATH_BAC120_MSA.format(prefix=prefix))
                marker_user_msa_path = os.path.join(
                    out_dir, PATH_BAC120_USER_MSA.format(prefix=prefix))
            else:
                self.logger.info(
                    'Processing %d genomes identified as archaeal.' % len(gids))
                marker_info_file = ar122_marker_info_file
                marker_filtered_genomes = os.path.join(
                    out_dir, PATH_AR122_FILTERED_GENOMES.format(prefix=prefix))
                marker_msa_path = os.path.join(
                    out_dir, PATH_AR122_MSA.format(prefix=prefix))
                marker_user_msa_path = os.path.join(
                    out_dir, PATH_AR122_USER_MSA.format(prefix=prefix))

            cur_genome_files = {
                gid: f for gid, f in genomic_files.items() if gid in gids}

            if skip_gtdb_refs:
                gtdb_msa = {}
            else:
                gtdb_msa = self._msa_filter_by_taxa(msa_file,
                                                    gtdb_taxonomy,
                                                    taxa_filter,
                                                    outgroup_taxon)
            gtdb_msa_mask = os.path.join(Config.MASK_DIR, mask_file)

            hmm_aligner = HmmAligner(self.cpus,
                                     self.pfam_top_hit_suffix,
                                     self.tigrfam_top_hit_suffix,
                                     self.protein_file_suffix,
                                     self.pfam_hmm_dir,
                                     self.tigrfam_hmms,
                                     Config.BAC120_MARKERS,
                                     Config.AR122_MARKERS,
                                     Config.RPS23_MARKERS)
            user_msa = hmm_aligner.align_marker_set(cur_genome_files,
                                                    marker_set_id)

            # filter columns without sufficient representation across taxa
            if skip_trimming:
                self.logger.info(
                    'Skipping custom filtering and selection of columns.')
                pruned_seqs = {}
                trimmed_seqs = merge_two_dicts(gtdb_msa, user_msa)

            elif custom_msa_filters:
                aligned_genomes = merge_two_dicts(gtdb_msa, user_msa)
                self.logger.info(
                    'Performing custom filtering and selection of columns.')

                trim_msa = TrimMSA(cols_per_gene,
                                   min_perc_aa / 100.0,
                                   min_consensus / 100.0,
                                   max_consensus / 100.0,
                                   min_per_taxa / 100.0,
                                   rnd_seed,
                                   os.path.join(out_dir, 'filter_%s' % marker_set_id))

                trimmed_seqs, pruned_seqs = trim_msa.trim(aligned_genomes,
                                                          marker_info_file)

                if trimmed_seqs:
                    self.logger.info('Filtered MSA from %d to %d AAs.' % (
                        len(list(aligned_genomes.values())[0]),
                        len(list(trimmed_seqs.values())[0])))

                self.logger.info('Filtered %d genomes with amino acids in <%.1f%% of columns in filtered MSA.' % (
                    len(pruned_seqs),
                    min_perc_aa))

                filtered_user_genomes = set(
                    pruned_seqs).intersection(user_msa)
                if len(filtered_user_genomes):
                    self.logger.info('Filtered genomes include %d user submitted genomes.' % len(
                        filtered_user_genomes))
            else:
                self.logger.info(
                    'Masking columns of multiple sequence alignment using canonical mask.')
                trimmed_seqs, pruned_seqs = self._apply_mask(gtdb_msa,
                                                             user_msa,
                                                             gtdb_msa_mask,
                                                             min_perc_aa / 100.0)
                self.logger.info('Masked alignment from %d to %d AAs.' % (len(list(user_msa.values())[0]),
                                                                          len(list(trimmed_seqs.values())[0])))

                if min_perc_aa > 0:
                    self.logger.info('%d user genomes have amino acids in <%.1f%% of columns in filtered MSA.' % (
                        len(pruned_seqs),
                        min_perc_aa))

            # write out filtering information
            with open(marker_filtered_genomes, 'w') as fout:
                for pruned_seq_id, pruned_seq in pruned_seqs.items():
                    if len(pruned_seq) == 0:
                        perc_alignment = 0
                    else:
                        valid_bases = sum(
                            [1 for c in pruned_seq if c.isalpha()])
                        perc_alignment = valid_bases * 100.0 / len(pruned_seq)
                    fout.write('%s\t%s\n' % (pruned_seq_id,
                                             'Insufficient number of amino acids in MSA (%.1f%%)' % perc_alignment))

            # write out MSAs
            if not skip_gtdb_refs:
                self.logger.info(
                    'Creating concatenated alignment for %d GTDB and user genomes.' % len(trimmed_seqs))
                self._write_msa(trimmed_seqs, marker_msa_path, gtdb_taxonomy)

            trimmed_user_msa = {
                k: v for k, v in trimmed_seqs.items() if k in user_msa}
            if len(trimmed_user_msa) > 0:
                self.logger.info(
                    'Creating concatenated alignment for %d user genomes.' % len(trimmed_user_msa))
                self._write_msa(trimmed_user_msa,
                                marker_user_msa_path, gtdb_taxonomy)
            else:
                if marker_set_id == 'bac120':
                    self.logger.info(
                        'All bacterial user genomes have been filtered out.')
                else:
                    self.logger.info(
                        'All archaeal user genomes have been filtered out.')

            # Create symlinks to the summary files
            if marker_set_id == 'bac120':
                symlink_f(PATH_BAC120_FILTERED_GENOMES.format(prefix=prefix),
                          os.path.join(out_dir, os.path.basename(PATH_BAC120_FILTERED_GENOMES.format(prefix=prefix))))
                if len(trimmed_user_msa) > 0:
                    symlink_f(PATH_BAC120_USER_MSA.format(prefix=prefix),
                              os.path.join(out_dir, os.path.basename(PATH_BAC120_USER_MSA.format(prefix=prefix))))
                if not skip_gtdb_refs:
                    symlink_f(PATH_BAC120_MSA.format(prefix=prefix),
                              os.path.join(out_dir, os.path.basename(PATH_BAC120_MSA.format(prefix=prefix))))
            elif marker_set_id == 'ar122':
                symlink_f(PATH_AR122_FILTERED_GENOMES.format(prefix=prefix),
                          os.path.join(out_dir, os.path.basename(PATH_AR122_FILTERED_GENOMES.format(prefix=prefix))))
                if len(trimmed_user_msa) > 0:
                    symlink_f(PATH_AR122_USER_MSA.format(prefix=prefix),
                              os.path.join(out_dir, os.path.basename(PATH_AR122_USER_MSA.format(prefix=prefix))))
                if not skip_gtdb_refs:
                    symlink_f(PATH_AR122_MSA.format(prefix=prefix),
                              os.path.join(out_dir, os.path.basename(PATH_AR122_MSA.format(prefix=prefix))))
            else:
                self.logger.error(
                    'There was an error determining the marker set.')
                raise GenomeMarkerSetUnknown



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

import multiprocessing as mp
import os
import sys
from collections import defaultdict

from gtdbtk.exceptions import GTDBTkExit
from gtdbtk.external.pypfam.Scan.PfamScan import PfamScan
from gtdbtk.tools import sha256


class PfamSearch(object):
    """Runs pfam_search.pl over a set of genomes."""

    def __init__(self,
                 threads,
                 pfam_hmm_dir,
                 protein_file_suffix,
                 pfam_suffix,
                 pfam_top_hit_suffix,
                 checksum_suffix,
                 output_dir):
        """Initialization."""

        self.threads = threads

        self.pfam_hmm_dir = pfam_hmm_dir
        self.protein_file_suffix = protein_file_suffix
        self.pfam_suffix = pfam_suffix
        self.pfam_top_hit_suffix = pfam_top_hit_suffix
        self.checksum_suffix = checksum_suffix
        self.output_dir = output_dir

    def _topHit(self, pfam_file):
        """Determine top hits to PFAMs.

        A gene may be assigned to multiple
        PFAM families from the same clan. The
        search_pfam.pl script takes care of
        most of these issues and here the results
        are simply parsed.

        Parameters
        ----------
        pfam_file : str
            Name of file containing hits to PFAM HMMs.
        """

        assembly_dir, filename = os.path.split(pfam_file)
        genome_id = filename.replace(self.pfam_suffix, '')
        output_tophit_file = os.path.join(self.output_dir, genome_id, filename.replace(self.pfam_suffix,
                                                                                       self.pfam_top_hit_suffix))

        tophits = defaultdict(dict)
        with open(pfam_file, 'r') as fh_pfam:
            for line in fh_pfam:
                if line[0] == '#' or not line.strip():
                    continue

                line_split = line.split()
                gene_id = line_split[0]
                hmm_id = line_split[5]
                evalue = float(line_split[12])
                bitscore = float(line_split[11])
                if gene_id in tophits:
                    if hmm_id in tophits[gene_id]:
                        if bitscore > tophits[gene_id][hmm_id][1]:
                            tophits[gene_id][hmm_id] = (evalue, bitscore)
                    else:
                        tophits[gene_id][hmm_id] = (evalue, bitscore)
                else:
                    tophits[gene_id][hmm_id] = (evalue, bitscore)

        with open(output_tophit_file, 'w') as fout:
            fout.write('Gene Id\tTop hits (Family id,e-value,bitscore)\n')
            for gene_id, hits in tophits.items():
                hit_str = []
                for hmm_id, stats in hits.items():
                    hit_str.append(hmm_id + ',' + ','.join(map(str, stats)))
                fout.write('%s\t%s\n' % (gene_id, ';'.join(hit_str)))

        # calculate checksum
        checksum = sha256(output_tophit_file)
        with open(output_tophit_file + self.checksum_suffix, 'w') as fout:
            fout.write(checksum)

    def _workerThread(self, queueIn, queueOut):
        """Process each data item in parallel."""
        try:
            while True:
                gene_file = queueIn.get(block=True, timeout=None)
                if gene_file is None:
                    break

                genome_dir, filename = os.path.split(gene_file)
                genome_id = filename.replace(self.protein_file_suffix, '')
                output_hit_file = os.path.join(self.output_dir, genome_id, filename.replace(self.protein_file_suffix,
                                                                                            self.pfam_suffix))

                pfam_scan = PfamScan(cpu=self.cpus_per_genome, fasta=gene_file, dir=self.pfam_hmm_dir)
                pfam_scan.search()
                pfam_scan.write_results(output_hit_file, None, None, None, None)

                # calculate checksum
                with open(output_hit_file + self.checksum_suffix, 'w') as fh:
                    fh.write(sha256(output_hit_file))

                # identify top hit for each gene
                self._topHit(output_hit_file)

                queueOut.put(gene_file)
        except Exception as error:
            raise error

    def _writerThread(self, numDataItems, writerQueue):
        """Store or write results of worker threads in a single thread."""
        processedItems = 0
        while True:
            a = writerQueue.get(block=True, timeout=None)
            if a is None:
                break

            processedItems += 1
            statusStr = '==> Finished processing %d of %d (%.1f%%) genomes.' % (processedItems,
                                                                                numDataItems,
                                                                                float(processedItems) * 100 / numDataItems)
            sys.stdout.write('\r%s' % statusStr)
            sys.stdout.flush()
        sys.stdout.write('\n')

    def run(self, gene_files):
        """Annotate genes with Pfam HMMs.

        Parameters
        ----------
        gene_files : iterable
            Gene files in FASTA format to process.
        """

        self.cpus_per_genome = max(1, int(self.threads / len(gene_files)))

        # populate worker queue with data to process
        workerQueue = mp.Queue()
        writerQueue = mp.Queue()

        for f in gene_files:
            workerQueue.put(f)

        for _ in range(self.threads):
            workerQueue.put(None)

        try:
            workerProc = [mp.Process(target=self._workerThread, args=(
                workerQueue, writerQueue)) for _ in range(self.threads)]
            writeProc = mp.Process(target=self._writerThread, args=(
                len(gene_files), writerQueue))

            writeProc.start()

            for p in workerProc:
                p.start()

            for p in workerProc:
                p.join()
                if p.exitcode != 0:
                    raise GTDBTkExit('An error was encountered while running hmmsearch.')

            writerQueue.put(None)
            writeProc.join()
        except Exception:
            for p in workerProc:
                p.terminate()

            writeProc.terminate()
            raise



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

import os
import re
import subprocess
import sys
from datetime import datetime

from ..HMM.HMMResults import HMMResults
from ..HMM.HMMResultsIO import HMMResultsIO
from ..HMM.HMMSequence import HMMSequence
from ..HMM.HMMUnit import HMMUnit


class PfamScan(object):
    """
    This class has been adapted from the Perl module written by Genome Research Ltd.
    Perl authors: Jaina Mistry (jm14@sanger.ac.uk), John Tate (jt6@sanger.ac.uk), Rob Finn (finnr@janelia.hhmi.org)
    Perl version: PfamScan.pm,v 1.4 2010-01-12 09:41:42 jm14 Exp $
    Python authors: Aaron Mussig (a.mussig@uq.edu.au)
    """

    def __init__(self, **kwargs):
        """
        The only constructor for the object. Accepts a set of arguments that specify the parameters for the search:
        cut_off, cir, clan_overlap, fasta, sequence, align, hmm, as
        """

        # Python adaptation: Initialise variables
        # Function: _read_pfam_data
        self._read = dict()

        # To avoid hard coding the location for the binary, we assume it will be on the path.....
        self._HMMSCAN = 'hmmscan'

        # handle arguments, if we were given any here
        self._process_args(kwargs)

    def _process_args(self, kwargs):
        """
        Handles the input arguments.
        """

        # make sure we get a sequence
        if 'fasta' in kwargs and 'sequence' in kwargs:
            sys.exit('FATAL: "-fasta" and "-sequence" are mutually exclusive')
        elif 'fasta' in kwargs:
            if not os.path.isfile(kwargs['fasta']):
                sys.exit('FATAL: fasta file "%s" doesn\'t exist' % kwargs['fasta'])
        elif 'sequence' in kwargs:
            if len(kwargs['sequence'] < 1):
                sys.exit('FATAL: no sequence given')
        else:
            sys.exit('FATAL: must specify either "-fasta" or "-sequence"')

        # check the cut off
        if ('e_seq' in kwargs and ('b_seq' in kwargs or 'b_dom' in kwargs)) or \
                ('b_seq' in kwargs and ('e_seq' in kwargs or 'e_dom' in kwargs)) or \
                ('b_dom' in kwargs and 'e_dom' in kwargs):
            sys.exit('FATAL: can\'t use e value and bit score threshold together')

        self._hmmscan_cutoff = list()
        if 'e_seq' in kwargs:
            sys.exit('FATAL: -e_seq functionality has not been implemented in Python')
        if 'e_dom' in kwargs:
            sys.exit('FATAL: -e_dom functionality has not been implemented in Python')
        if 'b_seq' in kwargs:
            sys.exit('FATAL: -b_seq functionality has not been implemented in Python')
        if 'b_dom' in kwargs:
            sys.exit('FATAL: -b_dom functionality has not been implemented in Python')

        if not self._hmmscan_cutoff:
            self._hmmscan_cutoff.append('--cut_ga')

        # make sure we have a valid directory for the HMM data files
        if not os.path.isdir(kwargs['dir']):
            sys.exit('FATAL: directory "%s" does not exist' % kwargs['dir'])

        # Populate the object
        self._cut_off = kwargs.get('cut_off', None)
        self._dir = kwargs.get('dir', None)
        self._clan_overlap = kwargs.get('clan_overlap', None)
        self._fasta = kwargs.get('fasta', None)
        self._align = kwargs.get('align', None)
        self._as = kwargs.get('as', None)
        self._sequence = kwargs.get('sequence', None)
        self._cpu = kwargs.get('cpu', None)
        self._translate = kwargs.get('translate', None)

        self._hmmlib = list()
        if 'hmmlib' in kwargs:
            if type(kwargs['hmmlib']) == list:
                [self._hmmlib.append(x) for x in kwargs['hmmlib']]
            else:
                self._hmmlib.append(kwargs['hmmlib'])
        else:
            self._hmmlib.append('Pfam-A.hmm')

        # Now check that the library exists in the data dir!
        for hmmlib in self._hmmlib:

            if not os.path.isfile(os.path.join(self._dir, hmmlib)) and \
                    os.path.isfile(os.path.join(self._dir, hmmlib + '.h3f')) and \
                    os.path.isfile(os.path.join(self._dir, hmmlib + '.h3i')) and \
                    os.path.isfile(os.path.join(self._dir, hmmlib + '.h3m')) and \
                    os.path.isfile(os.path.join(self._dir, hmmlib + '.h3p')) and \
                    os.path.isfile(os.path.join(self._dir, hmmlib + '.dat')):
                sys.exit('FATAL: can\'t find %s and/or %s binaries in "%s' % (hmmlib, hmmlib, kwargs['dir']))

            # read the necessary data, if it's not been read already
            self._read_pfam_data()

        self._max_seqname = 0

        # if there's nothing in "_sequence" try to load a fasta file
        if not self._sequence:
            self._read_fasta()

        # check again for a sequence. If we don't have one at this point, bail with
        # an error
        if not self._sequence:
            sys.exit('FATAL: no sequence given')

        # read fasta file, store maximum sequence name and store sequences for active
        # sites prediction
        if not self._max_seqname:
            self._parse_sequence()

        if self._as:
            sys.exit('FATAL: -as functionality has not been implemented in Python')

        if self._translate:
            sys.exit('FATAL: -translate functionality has not been implemented in Python')

        # see if a version number was specified
        self._version = kwargs.get('version', None)

    def _read_pfam_data(self):
        """
        Reads the Pfam data file ("Pfam-A.scan.dat") and populates the C<accmap>,
        C<nested> and C<clanmap> hashes on the object.
        """

        # print STDERR "reading " . $self->{_hmmlib} . ".dat\n" if($ENV{DEBUG});
        self._accmap = dict()
        self._nested = dict()
        self._clanmap = dict()
        self._desc = dict()
        self._seqGA = dict()
        self._domGA = dict()
        self._type = dict()
        self._model_len = dict()

        # Python adaptation: Pre-compile regex
        re_read_pfam_1 = re.compile(r'^\#=GF ID\s+(\S+)')
        re_read_pfam_2 = re.compile(r'^\#=GF\s+AC\s+(\S+)')
        re_read_pfam_3 = re.compile(r'^\#=GF\s+DE\s+(.+)')
        re_read_pfam_4 = re.compile(r'^\#=GF\s+GA\s+(\S+)\;\s+(\S+)\;')
        re_read_pfam_5 = re.compile(r'^\#=GF\s+TP\s+(\S+)')
        re_read_pfam_6 = re.compile(r'^\#=GF\s+ML\s+(\d+)')
        re_read_pfam_7 = re.compile(r'^\#=GF\s+NE\s+(\S+)')
        re_read_pfam_8 = re.compile(r'^\#=GF\s+CL\s+(\S+)')

        for hmmlib in self._hmmlib:
            scandat = '%s/%s.dat' % (self._dir, hmmlib)

            try:
                with open(scandat, 'r') as f:
                    SCANDAT = f.readlines()
            except IOError as e:
                sys.exit('FATAL: Couldn\'t open "%s" data file: %s' % (scandat, e.message))

            v_id = None
            for line in SCANDAT:

                # Python adaptation: Pre-search all regex patterns, as can't search and return in-place.
                res_read_pfam_1 = re_read_pfam_1.search(line)  # ^\#=GF ID\s+(\S+)
                res_read_pfam_2 = re_read_pfam_2.search(line)  # ^\#=GF\s+AC\s+(\S+)
                res_read_pfam_3 = re_read_pfam_3.search(line)  # ^\#=GF\s+DE\s+(.+)
                res_read_pfam_4 = re_read_pfam_4.search(line)  # ^\#=GF\s+GA\s+(\S+)\;\s+(\S+)\;
                res_read_pfam_5 = re_read_pfam_5.search(line)  # ^\#=GF\s+TP\s+(\S+)
                res_read_pfam_6 = re_read_pfam_6.search(line)  # ^\#=GF\s+ML\s+(\d+)
                res_read_pfam_7 = re_read_pfam_7.search(line)  # ^\#=GF\s+NE\s+(\S+)
                res_read_pfam_8 = re_read_pfam_8.search(line)  # ^\#=GF\s+CL\s+(\S+)

                if res_read_pfam_1:
                    v_id = res_read_pfam_1.group(1)
                elif res_read_pfam_2:
                    self._accmap[v_id] = res_read_pfam_2.group(1)
                elif res_read_pfam_3:
                    self._desc[v_id] = res_read_pfam_3.group(1)
                elif res_read_pfam_4:
                    self._seqGA[v_id] = float(res_read_pfam_4.group(1))
                    self._domGA[v_id] = float(res_read_pfam_4.group(2))
                elif res_read_pfam_5:
                    self._type[v_id] = res_read_pfam_5.group(1)
                elif res_read_pfam_6:
                    self._model_len[v_id] = int(res_read_pfam_6.group(1))
                elif res_read_pfam_7:
                    self._nested[v_id] = {res_read_pfam_7.group(1): 1}
                    self._nested[res_read_pfam_7.group(1)] = {v_id: 1}
                elif res_read_pfam_8:
                    self._clanmap[v_id] = res_read_pfam_8.group(1)

            # set a flag to show that we've read the data files already
            self._read[hmmlib] = True

    def _convert_results_search_to_scan(self, search_results):
        """
        Converts the search format to the scan format
        """

        scan_results = dict()

        for search_result in search_results:
            for seq_id in search_result.seqs:
                this_seq_obj = search_result.seqs[seq_id]

                if seq_id not in scan_results:
                    new_scan_result = HMMResults()
                    new_scan_result.seqName = this_seq_obj.name
                    new_scan_result.description = this_seq_obj.desc
                    new_scan_result.program = search_result.program

                    scan_results[seq_id] = new_scan_result

                this_scan_result = scan_results[seq_id]

                new_hmm_seq = HMMSequence()
                new_hmm_seq.evalue = this_seq_obj.evalue
                new_hmm_seq.bits = this_seq_obj.bits
                new_hmm_seq.bias = this_seq_obj.bias
                new_hmm_seq.exp = this_seq_obj.exp
                new_hmm_seq.numberHits = this_seq_obj.numberHits
                new_hmm_seq.name = search_result.seedName
                new_hmm_seq.desc = search_result.description
                this_scan_result.addHMMSeq(new_hmm_seq)

                for search_unit in this_seq_obj.hmmUnits:
                    new_hmm_unit = HMMUnit()
                    new_hmm_unit.name = search_result.seedName
                    new_hmm_unit.domain = search_unit.domain
                    new_hmm_unit.hmmalign = search_unit.hmmalign
                    new_hmm_unit.bits = search_unit.bits
                    new_hmm_unit.bias = search_unit.bias
                    new_hmm_unit.domEvalue = search_unit.domEvalue
                    new_hmm_unit.evalue = search_unit.evalue
                    new_hmm_unit.hmmFrom = search_unit.hmmFrom
                    new_hmm_unit.hmmTo = search_unit.hmmTo
                    new_hmm_unit.seqFrom = search_unit.seqFrom
                    new_hmm_unit.seqTo = search_unit.seqTo
                    new_hmm_unit.envFrom = search_unit.envFrom
                    new_hmm_unit.envTo = search_unit.envTo
                    new_hmm_unit.aliAcc = search_unit.aliAcc
                    this_scan_result.addHMMUnit(new_hmm_unit)

                this_scan_result.eof = True

        values = list()
        for key, value in sorted(scan_results.items()):
            values.append(value)

        return values

    def _read_fasta(self):
        """
        Reads a sequence from the fasta-format file that was specified in the parameters.
        """
        try:
            with open(self._fasta, 'r') as f:
                rows = f.readlines()
        except IOError as e:
            sys.exit('FATAL: Couldn\'t open fasta file "%s" %s' % (self._fasta, e.message))

        self._sequence_rows = rows
        self._sequence = ''.join(rows)

    def _parse_sequence(self):
        """
        This method is used to parse the sequence and hash it on sequence identifier.
        It also stores the length of the longest sequence id
        """
        seq_hash = dict()
        seq_id = None

        # Python adaptation: Pre-compile regex
        re_1 = re.compile(r'^\s*$')
        re_2 = re.compile(r'^>(\S+)')

        for row in self._sequence_rows:

            # Ignore blank lines
            if re_1.search(row):
                continue

            res_re_2 = re_2.search(row)
            if res_re_2:
                seq_id = res_re_2.group(1)

                if seq_id in seq_hash:
                    sys.exit('FATAL: Sequence identifiers must be unique. Your fasta file contains two sequences with the same id (%s)' % seq_id)

                # Store the max length of seq name, use this later when printing in ascii
                if self._max_seqname is None or len(seq_id) > self._max_seqname:
                    self._max_seqname = len(seq_id)

            else:
                if not seq_id:
                    sys.exit("FATAL: Unrecognised format of fasta file. Each sequence must have a header line in the format '>identifier  <optional description>'")
                row = row.rstrip()
                seq_hash[seq_id] = row

        self._seq_hash = seq_hash

    def search(self):
        """
        The main method on the object. Performs a C<hmmscan> search using the supplied sequence and the specified HMM library.
        """

        if not self._sequence:
            sys.exit('FATAL: no sequence given; set the search parameters before calling "search"')

        AllResults = dict()
        pfamB = None
        firstResult = None

        # Python adaptation: Pre-compile regex.
        re_1 = re.compile(r'Pfam\-B')

        for hmmlib in self._hmmlib:
            hmmscan_cut_off = list()
            seq_evalue = None
            dom_evalue = None
            if not re_1.search(hmmlib):
                hmmscan_cut_off = self._hmmscan_cutoff
            else:
                pfamB = True
                seq_evalue = 0.001
                dom_evalue = 0.001

                # It's a pfamB search so use some default cut off values
                hmmscan_cut_off.append('-E %s' % seq_evalue)
                hmmscan_cut_off.append('--domE %s' % dom_evalue)

            if self._cpu:
                params = ['hmmsearch', '--notextw', '--cpu', str(self._cpu),
                          ' '.join(hmmscan_cut_off),
                          os.path.join(self._dir, hmmlib), self._fasta]
            else:
                params = ['hmmsearch', '--notextw', ' '.join(hmmscan_cut_off),
                          os.path.join(self._dir, hmmlib), self._fasta]

            proc = subprocess.Popen(params, stdout=subprocess.PIPE,
                                    stderr=subprocess.STDOUT, encoding='utf-8')
            proc_out, proc_err = proc.communicate()
            proc_out_ascii = proc_out

            if proc_err:
                sys.exit('An error was encountered while running hmmsearch: %s' % proc_err)

            self._hmmresultIO = HMMResultsIO()
            self._all_results = self._hmmresultIO.parseMultiHMMER3(proc_out_ascii)
            self._all_results = self._convert_results_search_to_scan(self._all_results)

            if not re_1.search(hmmlib):

                if self._clan_overlap:
                    print('Resolve clan overlaps: off')
                else:
                    self._resolve_clan_overlap()

            # Determine which hits are significant
            for result in self._all_results:

                for unit in sorted(result.units, key=lambda x: x.seqFrom, reverse=False):
                    if not pfamB:
                        unit.sig = False
                        if result.seqs[unit.name].bits >= self._seqGA[unit.name]:
                            if unit.bits >= self._domGA[unit.name]:
                                unit.sig = True
            if firstResult:
                AllResults[self._all_results] = self._all_results
            else:
                firstResult = self._all_results

        # If more than one search, merge results into one object
        if len(AllResults) > 0:
            sys.exit('FATAL: multiple search functionality has not been implemented in Python')

        # print('<seq id> <alignment start> <alignment end> <envelope start> <envelope end> <hmm acc> <hmm name> <type> <hmm start> <hmm end> <hmm length> <bit score> <E-value> <significance> <clan>')

    def write_results(self, out, e_seq, e_dom, b_seq, b_dom):
        """
        Writes the results of the C<hmmscan> search. Takes a single argument, which can
        be an open filehandle or a filename. A fatal error is generated if a file of the
        given name already exists.
        """
        # if os.path.isfile(out):
        #     sys.exit('FATAL: output file "%s" already exists' % out)

        try:
            with open(out, 'w') as fh:

                # Python adaptation: Write the header.
                fh.write('#' * 84 + '\n')
                fh.write('# %s #\n' % "'||'''|,          '||'''|,  .|';                   ".center(80, ' '))
                fh.write('# %s #\n' % " ||   ||           ||   ||  ||                     ".center(80, ' '))
                fh.write('# %s #\n' % " ||...|' '||  ||`  ||...|' '||'   '''|.  '||),,(|, ".center(80, ' '))
                fh.write('# %s #\n' % " ||       `|..||   ||       ||   .|''||   || || || ".center(80, ' '))
                fh.write('# %s #\n' % ".||           ||  .||      .||.  `|..||. .||    ||.".center(80, ' '))
                fh.write('# %s #\n' % "           ,  |'                                   ".center(80, ' '))
                fh.write('# %s #\n' % "            ''                                     ".center(80, ' '))
                fh.write('# %s #\n' % ''.center(80, ' '))
                fh.write('# %s #\n' % 'This program is free software: you can redistribute it and/or modify'.center(80, ' '))
                fh.write('# %s #\n' % 'it under the terms of the GNU General Public License as published by'.center(80, ' '))
                fh.write('# %s #\n' % 'the Free Software Foundation, either version 3 of the License, or'.center(80, ' '))
                fh.write('# %s #\n' % '(at your option) any later version.'.center(80, ' '))
                fh.write('# %s #\n' % 'This program is distributed in the hope that it will be useful,'.center(80, ' '))
                fh.write('# %s #\n' % 'but WITHOUT ANY WARRANTY; without even the implied warranty of'.center(80, ' '))
                fh.write('# %s #\n' % 'MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the'.center(80, ' '))
                fh.write('# %s #\n' % 'GNU General Public License for more details.'.center(80, ' '))
                fh.write('# %s #\n' % 'You should have received a copy of the GNU General Public License'.center(80, ' '))
                fh.write('# %s #\n' % 'along with this program. If not, see <http://www.gnu.org/licenses/>.'.center(80, ' '))
                fh.write('# %s #\n' % ''.center(80, ' '))
                fh.write('# %s #\n' % 'PyPfam is a Python (lite) port of the PfamSearch scripts written in Perl.'.center(80, ' '))
                fh.write('# %s #\n' % 'Perl authors: Jaina Mistry (jm14@sanger.ac.uk), John Tate (jt6@sanger.ac.uk), '.center(80, ' '))
                fh.write('# %s #\n' % 'Rob Finn (rdf@sanger.ac.uk)'.center(80, ' '))
                fh.write('# %s #\n' % 'Python authors: Aaron Mussig (a.mussig@uq.edu.au)'.center(80, ' '))
                fh.write('# %s #\n' % ''.center(80, '='))
                fh.write('# %-25s %s\n' % ('Execution time:', datetime.now()))
                fh.write('# %-25s %s\n' % ('Query sequence file:', self._fasta))
                fh.write('# %-25s %s\n' % ('CPU threads:', self._cpu))
                fh.write('# %-25s %s\n' % ('Searching against:', os.path.join(self._dir, 'Pfam-A.hmm')))
                fh.write('# %-25s %s\n' % ('Resolve clan overlaps:', 'on'))
                fh.write('# %s #\n' % ''.center(80, '='))
                fh.write('# <seq id> <alignment start> <alignment end> <envelope start> <envelope end> <hmm acc> <hmm name> <type> <hmm start> <hmm end> <hmm length> <bit score> <E-value> <significance> <clan> \n')

                for result in self._all_results:
                    self._hmmresultIO.write_ascii_out(result, fh, self, e_seq, e_dom, b_seq, b_dom)

        except IOError as e:
            sys.exit('FATAL: Can\'t write to your output file "%s": %s' % (out, e.message))

    def _resolve_clan_overlap(self):
        """
        Resolves overlaps between clans.
        """

        no_clan_overlap = list()
        for result in self._all_results:
            new = result.remove_overlaps_by_clan(self._clanmap, self._nested)
            no_clan_overlap.append(new)

        self._all_results = no_clan_overlap



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

import logging
import os
import re
import subprocess
import sys

from gtdbtk.exceptions import PplacerException, TogException


class PplacerLogger(object):
    """Helper class for writing pplacer output."""

    def __init__(self, fh):
        """Initialise the class.

        Parameters
        ----------
        fh : BinaryIO
            The file to write to .
        """
        self.fh = fh

    def _disp_progress(self, line):
        """Calculates the progress and writes it to stdout.

        Parameters
        ----------
        line : str
            The line passed from pplacer stdout.
        """
        if not line.startswith('working on '):
            sys.stdout.write(f'\rInitialising pplacer [{line[0:50].center(50)}]')
            sys.stdout.flush()
        else:
            re_hits = re.search(r'\((\d+)\/(\d+)\)', line)
            current = int(re_hits.group(1))
            total = int(re_hits.group(2))
            sys.stdout.write('\r{}'.format(self._get_progress_str(current,
                                                                  total)))
            sys.stdout.flush()

    def _get_progress_str(self, current, total):
        """Determines the format of the genomes % string.

        Parameters
        ----------
        current : int
            The current number of genomes which have been placed.
        total : int
            The total number of genomes which are to be placed.

        Returns
        -------
        out : str
            A string formatted to show the progress of placement.
        """
        width = 50
        bar = str()
        prop = float(current) / total
        bar += '#' * int(prop * width)
        bar += '-' * (width - len(bar))
        return 'Placing genomes |{}| {}/{} ({:.2f}%)'.format(bar, current,
                                                             total, prop * 100)

    def read(self, line):
        """Reads a line and writes the progress to stdout and the file.

        Parameters
        ----------
        line : str
            A line returned from Prodigal stdout.
        """
        self.fh.write(line)
        line = line.strip()
        self._disp_progress(line)


class Pplacer(object):
    """Phylogenetic placement of genomes into a reference tree
    (http://matsen.fredhutch.org/pplacer/).
    """

    def __init__(self):
        """ Instantiate the class. """
        self.logger = logging.getLogger('timestamp')
        self.version = self._get_version()

    def _get_version(self):
        try:
            env = os.environ.copy()
            proc = subprocess.Popen(['pplacer', '--version'], stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE, env=env, encoding='utf-8')

            output, error = proc.communicate()
            return output.strip()
        except:
            return "(version unavailable)"

    def run(self, cpus, model, ref_pkg, json_out, msa_file, pplacer_out,
            mmap_file=None):
        """Place genomes into a reference tree.

        Args:
            cpus (int): The number of threads to use.
            model (str): The model to use. PROT: LG, WAG, JTT. NT: GTR.
            ref_pkg (str): The path to the reference package.
            json_out (str): The path to write the json output to.
            msa_file (str): The path to the input MSA file.
            pplacer_out (str): Where to write the pplacer output file.
            mmap_file (str, optional): The path to write a scratch file to.

        Raises:
            PplacerException: if a non-zero exit code, or if the json output
                              file isn't generated.

        """

        args = ['pplacer', '-m', model, '-j', str(cpus), '-c', ref_pkg, '-o',
                json_out, msa_file]

        if mmap_file:
            args.append('--mmap-file')
            args.append(mmap_file)

        proc = subprocess.Popen(args, stdout=subprocess.PIPE, encoding='utf-8')
        with open(pplacer_out, 'w') as fh:
            pplacer_logger = PplacerLogger(fh)
            while True:
                line = proc.stdout.readline()
                if not line:
                    sys.stdout.write('\n')
                    break
                pplacer_logger.read(line)
        proc.wait()

        if proc.returncode != 0:
            raise PplacerException('An error was encountered while '
                                   'running pplacer, check the log '
                                   'file: {}'.format(pplacer_out))

        if not os.path.isfile(json_out):
            self.logger.error('pplacer returned a zero exit code but no output '
                              'file was generated.')
            raise PplacerException

    def tog(self, pplacer_json_out, tree_file):
        """ Convert the pplacer json output into a newick tree.

        Args:
            pplacer_json_out (str): The path to the output of pplacer.
            tree_file (str): The path to output the newick file to.

        Raises:
            TogException: If a non-zero exit code is returned, or the tree file
                          isn't output.
        """

        args = ['guppy', 'tog', '-o', tree_file, pplacer_json_out]
        proc = subprocess.Popen(args, stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE)
        proc_out, proc_err = proc.communicate()

        if proc.returncode != 0:
            self.logger.error('An error was encountered while running tog.')
            raise TogException(proc_err)

        if not os.path.isfile(pplacer_json_out):
            self.logger.error('tog returned a zero exit code but no output '
                              'file was generated.')
            raise TogException
